// ----------------------------------------------------------------------------
//
//  Copyright (C) 2010-2014 Fons Adriaensen <fons@linuxaudio.org>
//    
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ----------------------------------------------------------------------------


#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <libgen.h>
#include "audiofile.h"
#include "config.h"



static int check_inout (unsigned int ip, unsigned int op)
{
    if (! size) return ERR_NOCONV;
    if ((ip < 1) || (ip > ninp)) return ERR_IONUM;
    if ((op < 1) || (op > nout)) return ERR_IONUM;
    return 0;
}


static int readfile (const char *line, int lnum, const char *cdir)
{
    int           n, err;
    unsigned int  i, out, nchan, nfram; 
    char          file [1024];
    char          path [1024];
    Audiofile     audio;
    float         gain;
    float        *buff;

    if (sscanf (line, "%u %f %n", &out, &gain, &n) != 2) return ERR_PARAM;
    n = sstring (line + n, file, 1024);
    if (!n) return ERR_PARAM;

    err = check_inout (1, out);
    if (err) return err;

    if (*file == '/') strcpy (path, file);
    else
    {
        strcpy (path, cdir);
        strcat (path, "/");
        strcat (path, file);
    }

    if (audio.open_read (path))
    {
        fprintf (stderr, "Line %d: Unable to open '%s'.\n", lnum, path);
        return ERR_OTHER;
    } 

    nchan = audio.chan ();
    nfram = audio.size ();

    if (audio.rate () != (int) fsamp)
    {
        fprintf (stderr, "Line %d: Warning: sample rate (%d) of '%s' does not match.\n",
                 lnum, audio.rate (), path);
    }

    if (nchan != ninp)
    {
	fprintf (stderr, "Line %d: Channel count (%d) of '%s' doens't match.\n",
                 lnum, nchan, path);
        audio.close ();
        return ERR_OTHER;
    } 

    if (nfram > size)
    {
	fprintf (stderr, "Line %d: Warning: data in '%s' will be truncated.\n",
                 lnum, path);
	nfram = size;
    } 

    try 
    {
        buff = new float [nfram * nchan];
    }
    catch (...)
    {
	audio.close ();
        return ERR_ALLOC;
    }

    audio.read (buff, nfram);
    for (i = 0; i < nchan; i++)
    {
	convproc->setimp (i, out - 1, gain, buff + i, nfram, nchan);
    }

    audio.close ();
    delete[] buff;
    return 0;
}


int config (const char *config)
{
    FILE          *F;
    int           stat, lnum;
    char          line [1024];
    char          cdir [1024];
    char          *p, *q;

    if (! (F = fopen (config, "r"))) 
    {
	fprintf (stderr, "Can't open '%s' for reading\n", config);
        return -1;
    } 

    strcpy (cdir, dirname ((char *) config));
    stat = 0;
    lnum = 0;

    while (! stat && fgets (line, 1024, F))
    {
        lnum++;
        p = line; 
        if (*p != '/')
	{
            while (isspace (*p)) p++;
            if ((*p > ' ') && (*p != '#'))
	    {
                stat = ERR_SYNTAX;
		break;
	    }
            continue;
	}
        for (q = p; (*q >= ' ') && !isspace (*q); q++);
        for (*q++ = 0; (*q >= ' ') && isspace (*q); q++);

        if (! strcmp (p, "/cd"))
        {
            if (sstring (q, cdir, 1024) == 0) stat = ERR_PARAM;
        }	
        else if (! strcmp (p, "/convolver/new"))   stat = convnew (q, lnum);
        else if (! strcmp (p, "/impfile/read"))    stat = readfile (q, lnum, cdir);
        else if (! strcmp (p, "/input/name"))      stat = inpname (q); 
        else if (! strcmp (p, "/output/name"))     stat = outname (q); 
        else stat = ERR_COMMAND;
    }

    fclose (F);
    if (stat == ERR_OTHER) stat = 0;
    if (stat)
    {
	fprintf (stderr, "Line %d: ", lnum);
	switch (stat)
	{
	case ERR_SYNTAX:
	    fprintf (stderr, "Syntax error.\n");
	    break;
	case ERR_PARAM:
	    fprintf (stderr, "Bad or missing parameters.\n");
	    break;
	case ERR_ALLOC:
	    fprintf (stderr, "Out of memory.\n");
	    break;
	case ERR_CANTCD:
	    fprintf (stderr, "Can't change directory to '%s'.\n", cdir);
	    break;
	case ERR_COMMAND:
	    fprintf (stderr, "Unknown command.\n");
	    break;
	case ERR_NOCONV:
	    fprintf (stderr, "No convolver yet defined.\n");
	    break;
	case ERR_IONUM:
	    fprintf (stderr, "Bad input or output number.\n");
	    break;
	default:
	    fprintf (stderr, "Unknown error.\n");		     
	}
    }

    return stat;
}

